import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import './web_socket_connection.dart';
import '../utils/config.dart';
import 'dart:convert';
import 'package:crypto/crypto.dart';
import 'package:http/http.dart' as http;

typedef void OnConnectCallback(String res);
typedef void OnRegisterCallback(Map<String, dynamic> res);
typedef void OnErrorCallback(int code, String reason);
typedef void OnLocalStreamCallback(MediaStream stream);
typedef void OnRemoteStreamCallback(MediaStream stream);
typedef void OnReceiveCallFromUserCallback(String receivefrom, String type);
typedef void OnAudioVideoStateInfoCallback(int audioFlag, int videoFlag);
typedef void OnCallRejectedByUserCallback();
typedef void OnCallAcceptedByUserCallback();
typedef void OnCallHungUpByUserCallback();
typedef void OnCallDeclineByYouCallback();
typedef void OnCallBusyCallback();

class SignalingClient {
  // IOWebSocketChannel _socket;
  static final SignalingClient _instance =
      SignalingClient._privateConstructor();
  static SignalingClient get instance => _instance;
  SignalingClient._privateConstructor();
  //callbacks
  OnConnectCallback onConnect;
  OnErrorCallback onError;
  OnRegisterCallback onRegister;
  OnLocalStreamCallback onLocalStream;
  OnRemoteStreamCallback onRemoteStream;
  OnReceiveCallFromUserCallback onReceiveCallFromUser;
  OnCallRejectedByUserCallback onCallRejectedByUser;
  OnCallAcceptedByUserCallback onCallAcceptedByUser;
  OnCallHungUpByUserCallback onCallHungUpByUser;
  OnCallDeclineByYouCallback onCallDeclineByYou;
  OnCallBusyCallback onCallBusyCallback;
  OnAudioVideoStateInfoCallback onAudioVideoStateInfo;

  WebSocketConnection _socket;
  RTCPeerConnection _peerConnection;
  MediaStream _localStream;
  MediaStream _audioStream;
  MediaStream _audioInternalStream;
  // MediaStream stream;
  Map<String, dynamic> requestIDType = {};
  String tenantID = "12345";
  String ref_id;
  String sessionUUID;
  bool screenShareflag = false;
  Stream sStream;
  static const MethodChannel _channel = const MethodChannel('flutter_vdotok');

  Future<dynamic> callAPI() async {
    final datarequest = {"api_key": api_key, "tenant_id": tenant_id};
    final response = await http.post(
      '$host',
      body: json.encode(datarequest),
    );
    print("this is response of Api call ${json.decode(response.body)}");
    if (response.statusCode == 200) {
      print("${response.statusCode}");
      return json.decode(response.body);
    } else {
      throw Exception("Failed to Load Data");
    }
  }

  Stream<int> numberStream() async* {
    // for(int i = 1; i<= 10; i++) {
    //   await Future.delayed(Duration(seconds: 1));

    yield 20;
    // }
    yield await _channel.invokeMethod('getInternalAudio', {"audio": true});
  }

  Future<void> connect() async {
    // Testing
    // _audioInternalStream   =
    //     await _channel.invokeMethod('getInternalAudio');
    // print("this is android native return ${_audioInternalStream.id}");
    // int errorCode = await MediaProjectionCreator.createMediaProjection();
    // if (errorCode == MediaProjectionCreator.ERROR_CODE_SUCCEED) {
    //   print("successed");
    // } else if (errorCode ==
    //     MediaProjectionCreator.ERROR_CODE_FAILED_USER_CANCELED) {
    //   print("Cancelled");
    // } else if (errorCode ==
    //     MediaProjectionCreator.ERROR_CODE_FAILED_SYSTEM_VERSION_TOO_LOW) {
    //   print("Failed: System API level need to higher than 21");
    // }
    var data = await callAPI();
    print("this is data ${data}");
    _socket =
        WebSocketConnection(url: data['media_server_map']['complete_address']);

    _socket.onOpen = () {
      print('onOpen');
      onConnect.call("connected");
    };

    _socket.onMessage = (message) {
      // print('Received data: ' + message);
      onMessage(json.decode(message));
    };

    _socket.onClose = (int code, String reason) {
      print("this is onclose socket");
      onError.call(code, reason);
      // print('Closed by server [$code => $reason]!');
      // onSignalingStateChange?.call(SignalingState.ConnectionClosed);
    };

    await _socket.connect();
  }

  closeSocket() {
    if (_socket != null) _socket.close();
  }

  void onMessage(message) async {
    Map<String, dynamic> mapData = message;
    print("this is data id ${mapData["requestType"]}");
    print("this is data- ${mapData}");

    switch (mapData["requestType"]) {
      //for register
      case 'register':
        if (mapData["responseCode"] == 200)
          onRegister.call(mapData);
        else
          onError.call(mapData["responseCode"], mapData["responseMessage"]);
        break;
      case '401':
        onError.call(mapData["responseCode"], mapData["responseMessage"]);
        break;
      //   // case 'peers':
      //   //   {
      //   //     // List<dynamic> peers = data;
      //   //     // if (onPeersUpdate != null) {
      //   //     //   Map<String, dynamic> event = Map<String, dynamic>();
      //   //     //   event['self'] = _selfId;
      //   //     //   event['peers'] = peers;
      //   //     //   onPeersUpdate?.call(event);
      //   //     // }
      //   //   }
      //   //   break;
      case 'incomingCall':
        {
          sessionUUID = mapData["sessionUUID"];
          incomingCall(mapData["from"], mapData["media_type"]);
        }
        break;
      case 'declineres':
        {
          onCallDeclineByYou.call();
          _peerConnection.close();
        }
        break;
      // startCommunication
      case 'startCommunication':
        {
          try {
            print("this is answer sdp ${mapData['sdpAnswer'].toString()}");
            _peerConnection
              ..setRemoteDescription(
                  RTCSessionDescription(mapData['sdpAnswer'], "answer"));
            // RTCSessionDescription(answerSdp, "offer"));
          } catch (e) {
            print("this is error $e");
          }
        }
        break;
      case 'stopCommunication':
        {
          onCallHungUpByUser.call();
          _closeSession();
        }
        break;
      case 'session_invite':
        print("this is data after accepting, ${mapData["responseCode"]}");
        if (mapData["responseCode"] == 100) {
          print("trying to locate & validate target");
        } else if (mapData["responseCode"] == 180) {
          print("trying to connect target");
        } else if (mapData["responseCode"] == 183) {
          print("target is alerting");
        } else if (mapData["responseCode"] == 407) {
          print("authentication required");
        } else if (mapData["responseCode"] == 404) {
          print("invalid target");
          onError.call(mapData["responseCode"], mapData["responseMessage"]);
          _closeSession();
        } else if (mapData["responseCode"] == 480) {
          print("Temporarily Unavailable");
        } else if (mapData["responseCode"] == 403) {
          print("Service not allowed");
        } else if (mapData["responseCode"] == 408) {
          print("No Response from target");
          onError.call(mapData["responseCode"], mapData["responseMessage"]);
          _closeSession();
        } else if (mapData["responseCode"] == 486) {
          print("Target Busy");
          onCallBusyCallback.call();
          _closeSession();
        } else if (mapData["responseCode"] == 496) {
          print("rejected by user");
          onCallRejectedByUser.call();
          _closeSession();
        } else if (mapData["responseCode"] == 200) {
          print("accepted");
        }
        break;
      case 'callResponse':
        {
          try {
            _peerConnection
              ..setRemoteDescription(
                  RTCSessionDescription(mapData['sdpAnswer'], "answer"));
            onCallAcceptedByUser.call();
          } catch (e) {
            print("this is error $e");
          }
        }
        break;
      case 'iceCandidate':
        {
          var candidateMap = mapData['candidate'];
          print(
              "this is icecandidate ${candidateMap['candidate']} ${candidateMap['sdpMid']} ${candidateMap['sdpMLineIndex']}");

          //       String jsonString = sdpController.text;
          // dynamic session = await jsonDecode('$jsonString');
          // print(session['candidate']);
          try {
            dynamic candidate = new RTCIceCandidate(candidateMap['candidate'],
                candidateMap['sdpMid'], candidateMap['sdpMLineIndex']);
            print("this is icecandidate $candidate");

            await _peerConnection?.addCandidate(candidate);
          } catch (e) {
            print("this is error $e");
          }

          // RTCIceCandidate candidate = RTCIceCandidate(candidateMap['candidate'],
          //     candidateMap['sdpMid'], candidateMap['sdpMLineIndex']);

          // if (session != null) {
          //   if (session.pc != null) {
          //     await session.pc.addCandidate(candidate);
          //   } else {
          //     session.remoteCandidates.add(candidate);
          //   }
          // } else {
          //   _sessions[sessionId] = Session(pid: peerId, sid: sessionId)
          //     ..remoteCandidates.add(candidate);
          // }
        }
        break;

      case "session_cancel":
        {
          onCallHungUpByUser.call();
          _closeSession();
        }
        break;

      case "state_information":
        {
          onAudioVideoStateInfo.call(
              mapData["audioInformation"], mapData["videoInformation"]);
          // onCallHungUpByUser.call();
          // _closeSession();
        }
        break;
      //   // case 'leave':
      //   //   {
      //   //     var peerId = data as String;
      //   //     _closeSessionByPeerId(peerId);
      //   //   }
      //   //   break;
      //   // case 'bye':
      //   //   {
      //   //     var sessionId = data['session_id'];
      //   //     print('bye: ' + sessionId);
      //   //     var session = _sessions.remove(sessionId);
      //   //     onCallStateChange?.call(session, CallState.CallStateBye);
      //   //     _closeSession(session);
      //   //   }
      //   //   break;
      //   // case 'keepalive':
      //   //   {
      //   //     print('keepalive response!');
      //   //   }
      //   //   break;
      default:
        break;
    }
    //in case of busy user
    // if (mapData["responseCode"] == 486) {
    //   print("this is decline");
    //   onCallBusyCallback.call();
    //   _closeSession();
    // }
  }

  register(dynamic user) {
    ref_id = user["ref_id"];
    Map<String, dynamic> registerjson = {
      "type": "request",
      "requestType": "register",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "tenantID": tenant_id,
      "referenceID": user["ref_id"],
      "authorizationToken": user["authorization_token"],
      "socketType": 0
    };

    requestIDType[registerjson["requestID"]] = "register";
    print("this is json $registerjson   $requestIDType");
    _socket.send(registerjson);
  }

  audioVideoState({int audioFlag, int videoFlag, String mcToken}) {
    print("this is statesssssss $audioFlag    $videoFlag");
    Map<String, dynamic> audioVideoStatejson = {
      "type": "request",
      "requestType": "state_information",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "referenceID": ref_id,
      "mcToken": mcToken,
      "sessionUUID": sessionUUID,
      "audioInformation": audioFlag,
      "videoInformation": videoFlag
    };
    print("this is state of audio and video json ${audioVideoStatejson}");

    _socket.send(audioVideoStatejson);
  }

  String generateMd5(String input) {
    return md5.convert(utf8.encode(input)).toString();
  }

  startCall(
      {String from,
      List<String> to,
      String mcToken,
      String meidaType,
      String callType,
      String sessionType}) async {
    screenShareflag = false;
    try {
      print("this is sdk $mcToken");
      _peerConnection = await _createPeerConnection(type: meidaType);
      _createOffer(
          from: from,
          to: to,
          mcToken: mcToken,
          mediaType: meidaType,
          callType: callType,
          sessionType: sessionType);
    } catch (e) {
      print("this is error $e");
    }
  }

  stopCall(mcToken) {
    Map<String, dynamic> jsonData = {
      "type": "request",
      "requestType": "session_cancel",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "sessionUUID": sessionUUID,
      "mcToken": mcToken
    };
    _socket.send(jsonData);
    // _peerConnection.close();
    _closeSession();
  }

  incomingCall(String name, String type) async {
    _peerConnection = await _createPeerConnection(type: type);
    onReceiveCallFromUser.call(name, type);
  }

  _createOffer(
      {String from,
      List<String> to,
      String mcToken,
      String mediaType,
      String callType,
      String sessionType}) async {
    print("this is create offer");
    try {
      RTCSessionDescription description = await _peerConnection.createOffer({
        "OfferToReceiveAudio": 1,
        "OfferToReceiveVideo": 1,
      });
      await _peerConnection.setLocalDescription(description);
      Map<String, dynamic> offerJson = {
        "from": from,
        "to": to,
        "type": "request",
        "requestType": "session_invite",
        "session_type": sessionType,
        "call_type": callType,
        "media_type": mediaType,
        "requestID": generateMd5(
            DateTime.now().millisecondsSinceEpoch.toString() +
                tenantID +
                ref_id),
        "sessionUUID": generateMd5(
            DateTime.now().millisecondsSinceEpoch.toString() +
                tenantID +
                ref_id),
        "mcToken": mcToken,
        "sdpOffer": description.sdp.toString()
      };
      requestIDType[offerJson["requestID"]] = "call";
      print("this is json on create offer $offerJson");
      sessionUUID = offerJson["sessionUUID"];
      // channel.sink.add(offerJson);
      _socket.send(offerJson);
    } catch (e) {
      print("this is error $e");
    }
  }

  createAnswer(name) async {
    print("this is name $name");
    try {
      print("this is peerConnection $_peerConnection");
      RTCSessionDescription description = await _peerConnection.createOffer({
        "OfferToReceiveAudio": 1,
        "OfferToReceiveVideo": 1,
      });
      // RTCSessionDescription description =
      //     await _peerConnection.createAnswer({'offerToReceiveVideo': 1});

      // var session = parse(description.sdp);
      // print(json.encode(session));
      // print(json.encode({
      //       'sdp': description.sdp.toString(),
      //       'type': description.type.toString(),
      //     }));

      _peerConnection.setLocalDescription(description);
      final Map<String, dynamic> _offerJson = {
        "type": "request",
        "requestID": generateMd5(
            DateTime.now().millisecondsSinceEpoch.toString() +
                tenantID +
                ref_id),
        "sessionUUID": sessionUUID,
        "requestType": "session_invite",
        "responseCode": 200,
        "responseMessage": "accepted",
        "sdpOffer": description.sdp.toString()
      };
      print("this is offer json $_offerJson");
      _socket.send(_offerJson);
      // channel.sink.add(_offerJson);
    } catch (e) {
      print("this is error $e");
    }
  }

  declineCall(from, mcToken) {
    Map<String, dynamic> jsonData = {
      "type": "response",
      "from": from,
      "requestType": "session_cancel",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "sessionUUID": sessionUUID,
      "mcToken": mcToken
      // "responseCode": 487,
      // "responseMessage": "rejected"
    };

    print("this is json data $jsonData");
    _socket.send(jsonData);
    _closeSession();
  }

  onCancelbytheCaller(mcToken) {
    Map<String, dynamic> jsonData = {
      "type": "request",
      "requestType": "session_cancel",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "sessionUUID": sessionUUID,
      "mcToken": mcToken
    };
    _socket.send(jsonData);
    _closeSession();
  }

  _createPeerConnection({String type}) async {
    try {
      Map<String, dynamic> configuration = {
        "iceServers": [
          {"url": "stun:stun.l.google.com:19302"},
        ]
      };

      final Map<String, dynamic> offerSdpConstraints = {
        "mandatory": {
          "OfferToReceiveAudio": true,
          "OfferToReceiveVideo": true,
        },
        "optional": [],
      };

      _localStream = await _getUserMedia(type: type);
      // _localStream?.getVideoTracks()[0].enabled = false;
      print("this is local stream ");

      onLocalStream.call(_localStream);

      RTCPeerConnection pc =
          await createPeerConnection(configuration, offerSdpConstraints);
      // var streams = pc.getLocalStreams();
      // for (var stream in streams) {
      //   _localStream = stream;
      // }
      onLocalStream.call(_localStream);
      pc.addStream(_localStream);
      pc.onIceCandidate = (e) {
        if (e.candidate != null) {
          Map<String, dynamic> icCandidate = {
            "requestType": "onIceCandidate",
            "type": "request",
            "referenceID": ref_id,
            "sessionUUID": sessionUUID,
            "candidate": {
              "candidate": "${e.candidate}",
              "sdpMid": "${e.sdpMid}",
              "sdpMLineIndex": e.sdpMlineIndex.toString()
            }
          };
          _socket.send(icCandidate);
        }
      };

      pc.onIceConnectionState = (e) {};
      // for remote strem

      pc.onAddStream = (stream) {
        print("this is remote stream ${stream.id}");
        onRemoteStream.call(stream);
      };

      return pc;
    } catch (e) {
      print("this is error $e");
    }
  }

  void switchCamera() {
    if (_localStream != null) {
      _localStream.getVideoTracks()[0].switchCamera();
    }
  }

  void enableCamera(enabled) {
    print("this is bool $enabled");
    if (_localStream != null) {
      // _localStream.getVideoTracks()[0].switchCamera();
      _localStream?.getVideoTracks()[0].enabled = enabled;
    }
  }

  muteMic() {
    if (_localStream != null) {
      bool enabled = _localStream.getAudioTracks()[0].enabled;
      _localStream.getAudioTracks()[0].enabled = !enabled;
      print("this is local ${_localStream.getAudioTracks()[0].enabled}");
      return !enabled;
    }
  }

  switchSpeaker(flag) {
    if (_localStream != null) {
      _localStream?.getAudioTracks()[0].enableSpeakerphone(flag);
    }
  }

  _getUserMedia({String type}) async {
    try {
      Map<String, dynamic> mediaConstraints = {
        'audio': false,
        'video': {
          // "height": {"min": "812", "ideal": "1920"},
          'mandatory': {
            //  'minWidth': '640',
            //   "maxWidth": '640',
            //   'minHeight': '1319',
            //   "maxHeight": '640',

            //   "minWidth":'640', // Provide your own width, height and frame rate here
            // "minHeight": '480',
            // "minFrameRate": '30',
          },
          'facingMode': 'user',
          // 'optional': [],
        }
      };
      Map<String, dynamic> onlyAudio = {'audio': true};

      _audioStream =
          await navigator.mediaDevices.getUserMedia(onlyAudio).catchError((e) {
        print("this is error audio  $e");
      });
      if (type == "video") {
        return await
            // navigator.getDisplayMedia(mediaConstraints)
            navigator.mediaDevices.getUserMedia(mediaConstraints).then((value) {
          print("this is stream ${value}");
          _audioStream.getAudioTracks().forEach((element) {
            value.addTrack(element);
          });
          if (!kIsWeb) value?.getAudioTracks()[0].enableSpeakerphone(true);
          return value;
        }).catchError((onError) {
          print("this is error $onError");
        });
      } else {
        if (!kIsWeb)
          _audioStream?.getAudioTracks()[0].enableSpeakerphone(false);
        return _audioStream;
      }

      if (screenShareflag) {
        // return await navigator.getDisplayMedia(mediaConstraints)
        //     // navigator.mediaDevices.getUserMedia(mediaConstraints)
        //     .then((value) {
        //   print("this is stream ${value}");
        //   // _audioStream.getAudioTracks().forEach((element) {
        //   //   value.addTrack(element);
        //   // });
        //   return value;
        // }).catchError((onError) {
        //   print("this is error $onError");
        // });
      } else {
        // return await
        //     // navigator.getDisplayMedia(mediaConstraints)
        //     navigator.mediaDevices.getUserMedia(mediaConstraints).then((value) {
        //   print("this is stream ${value}");
        //   _audioStream.getAudioTracks().forEach((element) {
        //     value.addTrack(element);
        //   });
        //   return value;
        // }).catchError((onError) {
        //   print("this is error $onError");
        // });
        // if (Platform.isAndroid) {
        //   if (_localStream == null) {
        //     // define initial variable here for ability to call stop() on each track

        //     return await navigator.mediaDevices
        //         .getDisplayMedia(mediaConstraints)
        //         .then((displayStream) async {
        //       audioStream.getAudioTracks().forEach((element) {
        //         displayStream.addTrack(element);
        //       });
        //       return displayStream;
        //       // return await displayStream
        //       //     .addTrack(audioStream.getTracks()[0]); // do stuff
        //       // or
        //       // do stuff
        //     }).catchError((e) {
        //       print("this is error over all $e");
        //     });

        //     // return await navigator.mediaDevices
        //     //     .getUserMedia(mediaConstraints)
        //     //     .then((value) {
        //     //   print("this is stream ${value.id}");
        //     //   audioStream.getAudioTracks().forEach((element) {
        //     //     value.addTrack(element);
        //     //   });
        //     //   return value;
        //     // }).catchError((onError) {
        //     //   print("this is error $onError");
        //     // });
        //   } else {
        //     return _localStream;
        //   }
        // } else {
        //   print("teseting ");
        //   // ReplayKitLauncher.launchReplayKitBroadcast('sharescreen');
        //   return await
        //       // navigator.mediaDevices.getDisplayMedia(mediaConstraints);
        //       navigator.mediaDevices
        //           .getUserMedia(mediaConstraints)
        //           .then((value) {
        //     audioStream.getAudioTracks().forEach((element) {
        //       value.addTrack(element);
        //     });
        //     print("this is stream ${value.id}");
        //     return value;
        //   }).catchError((onError) {
        //     print("this is error $onError");
        //   });
        // }
      }

      // }
      // } else {
      //   stream = await navigator.mediaDevices
      //       .getUserMedia(mediaConstraints)
      //       .then((value) => value)
      //       .catchError((onError) {
      //     print("this is error $onError");
      //   });
      // }

      // _localStream = stream;
      // _localRenderer.srcObject = stream;
      // _localRenderer.mirror = true;

      // _peerConnection.addStream(stream);

      // return stream;
    } catch (e) {
      print("this is error $e");
    }
  }

  Future<void> _closeSession() async {
    print("this is close session");
    _localStream?.getTracks()?.forEach((element) async {
      await element.stop();
    });
    await _localStream?.dispose();
    _localStream = null;

    await _peerConnection.close();

    _audioStream?.getTracks()?.forEach((element) async {
      await element.stop();
    });
    await _audioStream?.dispose();
    _audioStream = null;

    await _peerConnection.close();

    screenShareflag = false;
    sessionUUID = null;
    requestIDType.clear();
  }
}
